import base64
import gzip
import json
import os
import shutil
import unittest
import tempfile
# from datetime import timedelta
from datetime import datetime, timedelta
from io import StringIO
from typing import Tuple, List
import sys
from breeze_email_reports.EmailProfileReport import main

# REFERENCE_FILE_NAME = 'ReferenceData.json'
TEST_REFERENCE_WRONG_NAME = '2023-09-16T23:00:24.211429.json'
TEST_REFERENCE_NAME = '2023-09-17T23:00:24.211429.json'
TEST_REPLAY_NAME = '2023-10-17T23:00:24.211429.json'

TEST_FILES_DIR = os.path.join(os.path.split(__file__)[0], 'test_files')

# Data from previous run
TEST_PREVIOUS_DATA = os.path.join(TEST_FILES_DIR, 'ReferenceData.json')

# Returned by Breeze for "current" data
CURRENT_DATA_BREEZE = os.path.join(TEST_FILES_DIR, 'CurrentDataBreeze.json')
# Output from command, processed CURRENT_DATA_BREEZE. (Though this is uncompressed.)
TEST_CURRENT_DATA = os.path.join(TEST_FILES_DIR, 'ExpectedCurrentData.json')

# Expected results in email
EXPECTED_TABLE = os.path.join(TEST_FILES_DIR, 'ExpectedHtml.txt')
EXPECTED_CSV = os.path.join(TEST_FILES_DIR, 'ExpectedCSV.csv')
EXPECTED_TEXT = os.path.join(TEST_FILES_DIR, 'ExpectedText.txt')

TO_ADDRESS = 'to@test.com'
BCC_ADDRESS = 'bcc1@bcc.com, bcc2@bcc.com'


class MockBreezeAPI:
    def __init__(self, fields, profiles):
        self.fields = fields
        self.profiles = profiles

    def list_people(self, **kwargs):
        return self.profiles

    def get_profile_fields(self):
        return self.fields


class MockSender:
    def __init__(self):
        self.result = None

    def send_message(self, msg):
        self.result = msg


class TestSender(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.TemporaryDirectory()
        self.mock_sender = MockSender()
        self.saved_argv = sys.argv
        self.saved_stdout = None

    def runTest(self,
                report_format: str = 'text',
                do_gz: bool = True,
                do_reference: bool = True,
                previous_data: str = TEST_PREVIOUS_DATA,
                extra_params: List[str] = []) -> Tuple[str, int]:
        """
        Run BreezeProfile report. Return emitted event.
        :param report_format:
        :param do_gz:
        :param do_reference:
        :param previous_data:
        :param extra_params:
        :return:
        :raises: ParamException, IOException
        """

        with open(previous_data, 'r') as f:
            previous_data = f.read()

        test_file_name = os.path.join(self.test_dir.name, TEST_REFERENCE_NAME)
        if do_reference:
            if do_gz:
                with gzip.open(f'{test_file_name}.gz', 'w') as outfile:
                    outfile.write(bytes(previous_data, 'utf8'))
            else:
                with open(test_file_name, 'w') as f:
                    f.write(previous_data)
            with open(os.path.join(self.test_dir.name, TEST_REFERENCE_WRONG_NAME),
                      'w') as f:
                f.write('This is bad data')


        with open(CURRENT_DATA_BREEZE, 'r') as f:
            fields, profiles = json.load(f)

        api_mock = MockBreezeAPI(fields, profiles)

        sys.argv = ['test', '-f', 'from@test.com', '-t', TO_ADDRESS,
                    '-b', BCC_ADDRESS,
                    '--data', self.test_dir.name,
                    '--log_level=critical',
                    f'--logfile={os.path.join(self.test_dir.name, "test.log")}',
                    '--format', report_format] + extra_params

        ret = main(breeze_api=api_mock, email_sender=self.mock_sender)
        self.assertEqual(None, ret)
        # files = os.listdir(self.test_dir.name)
        # newfile = files[-1]
        # with open(os.path.join(self.test_dir.name, newfile), 'r') as f:
        #     expect_out = f.read()
        #
        # with open(TEST_CURRENT_DATA, 'w') as f:
        #     f.write(expect_out)

    def test_html(self):
        self.runTest('html',
                     extra_params=['--reference_data', TEST_PREVIOUS_DATA])

        sent_message = self.mock_sender.result
        # daw30410@yahoo.com
        # bcc1@foo,bcc2@foo
        # Breeze profile change report
        # [<email.mime.text.MIMEText object at 0x10e514cd0>, <email.mime.text.MIMEText object at 0x10edb8390>]
        # ['Content-Type', 'MIME-Version', 'Content-Transfer-Encoding']
        # ['text/plain; charset="us-ascii"', '1.0', '7bit']
        # ['Content-Type', 'MIME-Version', 'Content-Transfer-Encoding']
        # ['text/html; charset="us-ascii"', '1.0', '7bit']
        # Breeze changes since Sep 17 2023 11:00PM (Payload one get_payload()
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(str(pl0.get_payload()).
                        startswith('Breeze changes for Unknown thru'))
        self.assertEqual(pl1.get_content_subtype(), 'html')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TABLE, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TABLE, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_csv(self):
        self.runTest('csv', do_gz=True)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes since Sep 17 2023 11:00PM thru')))
        self.assertEqual(pl1.get_content_subtype(), 'csv')
        got_bytes = base64.b64decode(pl1.get_payload())
        got = got_bytes.decode('utf-8)')
        # with open(EXPECTED_CSV, 'w', newline='') as f:
        #     f.write(got)
        with open(EXPECTED_CSV, 'r', newline='') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_replay(self):
        shutil.copyfile(TEST_CURRENT_DATA,
                        os.path.join(self.test_dir.name, TEST_REPLAY_NAME))
        self.runTest('html', extra_params=['--replay'], do_gz=False)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertEqual(
            'Breeze changes for Sep 17 2023 11:00PM thru Oct 17 2023 11:00PM',
            str(pl0.get_payload()))
        # self.assertTrue(str(pl0.get_payload()).
        #                 startswith('Breeze changes for Unknown thru'))
        self.assertEqual(pl1.get_content_subtype(), 'html')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TABLE, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TABLE, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_no_change(self):
        self.runTest('csv', previous_data=TEST_CURRENT_DATA)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))
        payloads = sent_message.get_payload()
        self.assertEqual(len(payloads), 1)
        pl0 = payloads[0]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('No changes found between Sep 17 2023 11:00PM thru ')))

    def test_text(self):
        self.runTest('text')
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes for Sep 17 2023 11:00PM')))
        self.assertEqual(pl1.get_content_subtype(), 'plain')
        got = str(pl1.get_payload())
        # with open(EXPECTED_TEXT, 'w') as f:
        #     f.write(got)
        with open(EXPECTED_TEXT, 'r') as f:
            ref = f.read()
        self.assertEqual(got, ref)

    def test_bad_params(self):
        # Exception if no "to"
        sys.argv = ['test', '-b', 'foo@bar']
        with self.assertRaises(SystemExit) as se:
            main()
        self.assertEqual('--from=sender is required', se.exception.code)

    def test_no_to(self):
        sys.argv = ['test', '-f', 'from@foobar', ]
        with self.assertRaises(SystemExit) as se:
            main()
        self.assertEqual('Either -t or -b is required', se.exception.code)

    def test_bad_data(self):
        sys.argv = ['test', '-f', 'from@foobar', '-t', 'to@foobar',
                    '--logfile=stdout',
                    '--log_level=critical',
                    '--data', '/no/such/dir']
        with self.assertRaises(SystemExit) as e:
            main(email_sender=self.mock_sender)
        self.assertEqual('Directory /no/such/dir doesn\'t exist',
                         e.exception.code)

    def test_no_previous(self):
        self.runTest(report_format='text', do_reference=False)
        sent_message = self.mock_sender.result
        self.assertEqual('Breeze profile change report',
                         str(sent_message.get('Subject')))
        self.assertEqual(TO_ADDRESS, sent_message.get('To'))
        self.assertEqual(BCC_ADDRESS, sent_message.get('Bcc'))

        # sent_message = self.mock_sender.result
        payloads = sent_message.get_payload()
        pl0 = payloads[0]
        pl1 = payloads[1]
        self.assertEqual(pl0.get_content_subtype(), 'plain')
        self.assertTrue(
            str(pl0.get_payload().
                startswith('Breeze changes for Unknown thru')))
        self.assertEqual(pl1.get_content_subtype(), 'plain')
        # got = str(pl1.get_payload())
        # print(got)
        # with open(EXPECTED_TEXT, 'w') as f:
        #     f.write(got)

    def test_retain(self):
        now = datetime.now()
        days = 7
        files = [datetime.isoformat(now - timedelta(days=days + d)) + '.json'
                 for d in [5, 1, -1, -2]]
        for f in files:
            shutil.copyfile(TEST_CURRENT_DATA, os.path.join(self.test_dir.name, f))

        self.runTest(report_format='text',
                     do_reference=False,
                     extra_params=['--retain_days', str(days)])
        retained_files = os.listdir(self.test_dir.name)
        self.assertEqual(3, len(retained_files))
        self.assertFalse(files[1] in retained_files)
        self.assertTrue(files[3] in retained_files)

    def test_list_directories(self):
        sys.argv = ['test', '--list_directories', '--log_level=critical']
        self.saved_stdout = sys.stdout
        capture_out = StringIO()
        sys.stdout = capture_out
        with self.assertRaises(SystemExit) as e:
            main(email_sender=self.mock_sender)
        self.assertEqual(e.exception.code, 0,
                         "List directories shouldn't cause error")
        sys.stdout = self.saved_stdout
        capture_out.seek(0)
        output = capture_out.read()
        lines = output.split('\n')
        self.assertEqual('configured_mail_sender configuration files:',
                         lines[0])

    def test_known_domains(self):
        sys.argv = ['test', '--list_domains', '--log_level=critical']
        capture_out = StringIO()
        self.saved_stdout = sys.stdout
        sys.stdout = capture_out
        with self.assertRaises(SystemExit) as e:
            main()
        self.assertEqual(e.exception.code, 0,
                         "List domains shouldn't cause error")
        sys.stout = self.saved_stdout
        capture_out.seek(0)
        captured = capture_out.read()
        # lines = captured.split('\n')
        domains = {}
        for line in captured.split('\n')[1:]:
            line = line.replace(' ', '')
            if line:
                entry = line.split(':')
                key = entry[0]
                domains[key] = entry[1]
        self.assertEqual('smtp.mail.yahoo.com', domains.get('yahoo.com'))

    def test_initialize(self):
        os.removedirs(self.test_dir.name)
        with self.assertRaises(SystemExit) as e:
            self.runTest(extra_params=['--initialize'],
                         do_reference=False)
        self.assertEqual(0, e.exception.code)
        self.assertIsNone(self.mock_sender.result)
        # Should get same result
        files = os.listdir(self.test_dir.name)
        self.assertEqual(1, len(files))
        result = files[0]
        path = os.path.join(self.test_dir.name, result)
        with gzip.open(path, 'r') as f:
            saved_data = f.read()
        with open(TEST_CURRENT_DATA, 'r') as f:
            expected = bytes(f.read(), 'utf-8')
        self.assertEqual(expected, saved_data)

    def tearDown(self):
        if self.saved_stdout:
            sys.stdout = self.saved_stdout
        self.test_dir.cleanup()
        sys.argv = self.saved_argv



import os
from io import StringIO
from breeze_email_reports.table_format import (_HTMLFormatter, ColSpec,
                                               _splitval, _TextFormatter,
                                               _CSVFormatter)
import unittest

TEST_DIR = os.path.join(os.path.split(__file__)[0], 'split_test_files')
# TEST_FILES_DIR = os.path.join(TEST_DIR, 'test_files')
EXPECT_TEXT_FILE = os.path.join(TEST_DIR, 'ExpectedTableText.txt')
EXPECT_HTML_FILE = os.path.join(TEST_DIR, 'ExpectedTableHTML.html')
EXPECT_CSV_FILE = os.path.join(TEST_DIR, 'ExpectedTableCSV.csv')


class TestSplitVal(unittest.TestCase):
    def test_splitVal(self):
        # c0_width = 15
        # c1_width = 20
        # c1_width = 10
        # cs0 = ColSpec('c0', width=15)
        # cs1 = ColSpec('c1', width=20)
        cs2 = ColSpec('c2', width=10)

        v0 = 'abcdefghij'
        s0, s1 = _splitval(v0, cs2)
        self.assertEqual(s0, v0)
        self.assertEqual(s1, '')
        s0, s1 = _splitval(v0 + 'xy', cs2)
        self.assertEqual(s0, v0)
        self.assertEqual(s1, 'xy')

        v1 = 'abcd:ef ghij'
        s0, s1 = _splitval(v1, cs2)
        self.assertEqual(s0, 'abcd:')
        self.assertEqual(s1, 'ef ghij')

        s0, s1 = _splitval(v1, cs2, splits=' ')
        self.assertEqual(s0, 'abcd:ef')
        self.assertEqual(s1, 'ghij')


class TestFormatters(unittest.TestCase):
    def setUp(self):
        self.cs0 = ColSpec('c0', width=30)
        self.cs1 = ColSpec('c1', width=20)
        self.cs2 = ColSpec('c2', width=20)
        self.col_specs = [self.cs0, self.cs1, self.cs2]
        self.data = [('aname',
                      [
                        [
                            'afield',
                            'oldval',
                            'newval'
                        ],
                        [
                            'notherfield:withbreak',
                            None,
                            ['new1', 'new2 with a long value']
                        ]
                      ]
                     )
                    ]

    def test_text_format(self):
        formatter = _TextFormatter(self.col_specs)

        out = StringIO()
        formatter.format_table(self.data, out)
        out.seek(0)
        result = out.read()
        # with open(EXPECT_TEXT_FILE, 'w', newline='') as f:
        #     f.write(result)
        with open(EXPECT_TEXT_FILE, 'r', newline='') as f:
            expect = f.read()
        self.assertEqual(expect, result)

    def test_html_format(self):

        formatter = _HTMLFormatter(self.col_specs)

        out = StringIO()
        formatter.format_table(self.data, out)
        out.seek(0)
        result = out.read()
        # with open(EXPECT_HTML_FILE, 'w', newline='') as f:
        #     f.write(result)
        with open(EXPECT_HTML_FILE, 'r', newline='') as f:
            expect = f.read()
        self.assertEqual(expect, result)

    def test_csv_format(self):
        formatter = _CSVFormatter(self.col_specs)

        out = StringIO()
        formatter.format_table(self.data, out)
        out.seek(0)
        result = out.read()
        # with open(EXPECT_CSV_FILE, 'w', newline='') as f:
        #     f.write(result)
        with open(EXPECT_CSV_FILE, 'r', newline='') as f:
            expect = f.read()
        self.assertEqual(expect, result)




import csv
import html
from abc import abstractmethod
from io import IOBase
from typing import List, Tuple, Union

# Pad string to length, e.g. f'{a<15}' will pad to 15 chars.


class ColSpec:
    def __init__(self,
                 title: str,
                 header_attrs: str = '',
                 row_attrs: str = '',
                 width: int = 0):
        """
        Column specifications for table formatter
        :param title: Column title
        :param header_attrs: HTML attributes for header (if appropriate)
        :param row_attrs: HTML attributes for data rows (if appropriate)
        """
        self.title_val = title
        self.header_attrs_val = header_attrs if header_attrs else row_attrs
        self.row_attrs_val = row_attrs if row_attrs else header_attrs
        self.width_val = width

    @property
    def title(self):
        return self.title_val

    @property
    def header_attrs(self):
        return self.header_attrs_val

    @property
    def row_attributes(self):
        return self.row_attrs_val

    @property
    def width(self):
        return self.width_val


class _ReportFormatter:
    def __init__(self,
                 column_specs: List[ColSpec]):
        self.column_specs = column_specs
        self.columns = len(column_specs)

    @abstractmethod
    def format_table(self,
                     data: List[Tuple],
                     output: IOBase):
        """
        Format the data in appropriate format and write to output
        :param data: Data for the table
        :param output: Output stream to write to.
        :return: None
        The input data is a list of tuples. The first value is a group header
        that divides subsequent data. (A member name in this context.) The second
        value is a list of tuples, each with label for the tuple followed by values
        for columns in the row. Each column value can be None, a string, or a list
        of strings.
        """
        pass


def _escape_html_value(val: Union[str, List[str]]) -> str:
    if val:
        if isinstance(val, list):
            return '<br/>'.join([html.escape(v).replace(';', '<br/>') for v in val])
        else:
            return html.escape(val)
    else:
        return ' '


class _HTMLFormatter(_ReportFormatter):
    def __init__(self, *args, **kwargs):
        _ReportFormatter.__init__(self, *args, **kwargs)
        self.table_header = ('<table border="1" style="border:1px solid #000000;'
                             'border-collapse:collapse" cellpadding="4">\n')

    def format_table(self,
                     data: List[Tuple],
                     output: IOBase):
        """
        Format as an HTML table
        :param data:
        :param output:
        :return:
        """
        output.write('<table border="1" style="border:1px solid #000000;'
                     'border-collapse:collapse" cellpadding="4">')
        nl = '\n'
        row = '\n'.join([f'  <th {c.header_attrs}>{c.title}</th>'
                         for c in self.column_specs])
        output.write(f'{nl} <tr>{nl}{row}{nl} </tr>{nl}')

        for name, values in data:
            output.write(f'<tr><td colspan="{self.columns}" '
                         f'bgcolor="LightBlue"><b>{html.escape(name)}</b></td></tr>')
            for row in values:
                output.write(' <tr>\n')
                for i in range(len(row)):
                    cs = self.column_specs[i]
                    colval = row[i]
                    output.write(f'  <td {cs.row_attributes}>'
                                 f'{_escape_html_value(colval)}</td>' '\n')
                output.write(' </tr>\n')

        output.write('</table>\n')


def _splitval(val: str, cs: ColSpec, splits: str = ';: ') \
        -> Tuple[str, Union[str, None]]:
    """
    Split a string if too long
    :param val: Value to split
    :param cs: ColSpec with desired length
    :return: Tuple, parts of val before and after the split
    """
    # use .strip() to remove pre and post whitespace
    if not val:
        return '', None
    if len(val) <= cs.width:
        return val, ''
    last_found = cs.width - 1
    found_char = None
    for c in splits:
        end = cs.width if c == ' ' else cs.width - 1
        found = val.rfind(c, 0, end)
        if found > 0:
            last_found = found
            found_char = c
            break

    first = val[:last_found] if found_char == ' ' else val[:last_found + 1]
    second = val[last_found + 1:]
    return first, second


class _ToFileFormatter(_ReportFormatter):
    def __init__(self, *args, **kwargs):
        _ReportFormatter.__init__(self, *args, **kwargs)
        # self.line_mark = '   |'
        # self.row_prefix = self.line_mark
        # for cs in self.column_specs:
        #     self.line_mark += ''.rjust(cs.width, '-') + '|'

    def format_table(self,
                     data: List[Tuple],
                     output: IOBase):
        """
        Format as an HTML table
        :param data:
        :param output:
        :return:
        """
        for section in data:
            name, values = section
            self.start_section(name, output)
            for row in values:
                more = True
                self.start_group(output)
                # print(self.line_mark, file=output)
                while more:
                    # print(self.row_prefix, end='', file=output)
                    more = False
                    next_values = []
                    this_row = []
                    for i in range(min(len(row), self.columns)):
                        cs = self.column_specs[i]
                        colval = row[i]
                        nextval = []
                        if colval:
                            if isinstance(colval, list):
                                curval = colval[0]
                                nextval = colval[1:]
                                if nextval:
                                    more = True
                            else:
                                curval = colval
                                nextval = []
                            curval, overflow = self.splitval(curval, cs)
                            if overflow:
                                nextval = [overflow] + nextval
                                more = True
                        else:
                            curval = ''
                        this_row.append(curval if curval else '')
                        next_values.append(nextval)

                    self.emit_row(this_row, output)
                    row = next_values

            # print(self.line_mark, file=output)
            self.end_section(output)
        self.finish_output(output)

    @abstractmethod
    def start_section(self, name: str, output: IOBase):
        pass

    def end_section(self, output: IOBase):
        pass

    def start_group(self, output):
        pass

    def splitval(self, val: str, cs: ColSpec) -> Tuple[str, Union[str, None]]:
        return val, None

    def emit_row(self, values: List[str], output: IOBase):
        # print(self.row_prefix, end='', file=output)
        # for i in range(len(values)):
        #     v = values[i]
        #     cs = self.column_specs[i]
        #     v = v.ljust(cs.width)
        #     print(v + '|',
        #           end='',
        #           file=output)
        #     output.flush()
        # print('', file=output)
        pass

    def finish_output(self, output: IOBase):
        pass


class _TextFormatter(_ToFileFormatter):
    def __init__(self, *args, **kwargs):
        _ToFileFormatter.__init__(self, *args, **kwargs)
        self.line_mark = '   |'
        self.row_prefix = self.line_mark
        for cs in self.column_specs:
            self.line_mark += ''.rjust(cs.width, '-') + '|'

    def start_section(self, name: str, output: IOBase):
        print('', file=output)
        print(name, file=output)

    def start_group(self, output):
        print(self.line_mark, file=output)

    def splitval(self, val: str, cs: ColSpec) -> Tuple[str, str]:
        ret_val, overflow = _splitval(val, cs)
        return ret_val.replace('\n', ';'), ('+' + overflow) if overflow else None

    def emit_row(self, values: List[str], output: IOBase):
        print(self.row_prefix, end='', file=output)
        for i in range(len(values)):
            v = values[i]
            cs = self.column_specs[i]
            v = v.ljust(cs.width)
            print(v + '|',
                  end='',
                  file=output)
            output.flush()
        print('', file=output)

    def end_section(self, output: IOBase):
        print(self.line_mark, file=output)


class _CSVFormatter(_ToFileFormatter):
    def __init__(self, *args, **kwargs):
        _ToFileFormatter.__init__(self, *args, **kwargs)
        self.writer = None

    def start_section(self, name: str, output: IOBase):
        if not self.writer:
            self.writer = csv.writer(output)
        self.writer.writerow([name])

    def emit_row(self, values: List[str], output: IOBase):
        self.writer.writerow([''] + values)



from io import StringIO
from pathlib import Path
from typing import List, Tuple, Dict, Union

import configured_mail_sender
from configured_mail_sender import create_sender, MailSender, known_domains
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from breeze_chms_api.profile_helper import ProfileHelper, join_dicts, profile_compare
from breeze_chms_api import breeze

import logging
import sys
import json
import gzip
import argparse
import os
import platformdirs
import datetime

from breeze_email_reports.table_format import ColSpec, _HTMLFormatter, _CSVFormatter, _TextFormatter

DEFAULT_DATA_DIR = platformdirs.user_data_dir('BreezeProfiles')
DEFAULT_COLUMN_WIDTHS = '30,20,20'

APPLICATION_NAME = 'email_profile_report'
DEFAULT_LOG_FILE = os.path.join(platformdirs.user_log_dir(), APPLICATION_NAME + '.log')


class ProfileData:
    def __init__(self):
        self.datetime_str = 'Unknown'
        self.fields_map = {}
        self.profiles = {}

    def get_datetime(self) -> str:
        """
        Get time of this data dump
        :return: Date time in iso format
        """
        return self.datetime_str

    def get_fields_map(self) -> Dict[str, Dict[str, Union[str, List[str]]]]:
        """
        Get mapping from field id to name
        :return: The mapping
        """
        return self.fields_map

    def get_profile_fields(self) -> Dict[str, Dict[str, Union[str, List[str]]]]:
        """
        Return profile field values as list of tuples. First item in each tuple
        is the profile id. Second is a mapping from field id to Value(s) of
        the field.
        :return: List of profiles with their field values
        """
        return self.profiles

    def save_data(self) -> None:
        """
        Save current data to a new file in the given directory,
        if this was freshly loaded from Breeze.
        :return: None
        """
        return


class ProfileFromBreeze(ProfileData):
    """
    Profile data fetched directly from Breeze.
    """

    def __init__(self,
                 args,
                 breeze_api: breeze.BreezeApi = None,
                 do_save: bool = True):
        """
        Get current data from breeze
        :param args: Command line arguments
        :param breeze_api: Pre-existing Breeze API
        :param do_save: Save results on completion
        """
        ProfileData.__init__(self)
        self.do_save = do_save
        self.args = args
        ProfileData.__init__(self)
        breeze_api = breeze_api if breeze_api \
            else breeze.breeze_api(overrides=args.breeze_creds)
        profile_helper = ProfileHelper(breeze_api.get_profile_fields())
        self.fields_map = profile_helper.get_field_id_to_name()
        people = breeze_api.list_people(details=True)
        self.profiles = profile_helper.process_profiles(people)
        curtime = datetime.datetime.now()
        self.save_file_name = (datetime.datetime.isoformat(curtime)
                               + '.json.gz')
        self.datetime_str = curtime.strftime("%b %d %Y %I:%M%p")

    def save_data(self) -> None:
        if self.do_save:
            new_file = os.path.join(self.args.data, self.save_file_name)
            with gzip.open(new_file, 'w') as outfile:
                towrite = json.dumps((self.fields_map, self.profiles), indent=2)
                outfile.write(bytes(towrite, 'utf8'))
            if self.args.retain_days > 0:
                # We need to clean up old files
                oldest_allowed = (datetime.datetime.now()
                                  - datetime.timedelta(days=self.args.retain_days)
                                  ).isoformat()
                now_files = _get_previous_files(self.args)
                too_old = [f for f in now_files
                           if f < oldest_allowed]
                for file in too_old:
                    path = os.path.join(self.args.data, file)
                    logging.info(f'Removing old data: {file}')
                    os.remove(path)


class ProfileFromFile(ProfileData):
    """
    Profile data loaded from a file.
    """

    def __init__(self, filepath: str):
        """
        Load profile fields from a file
        :param filepath:
        """
        ProfileData.__init__(self)
        directory, file = os.path.split(filepath)
        if filepath.endswith('.gz'):
            date_part = file[:-8]
            with gzip.open(filepath, 'r') as prfile:
                data = json.loads(prfile.read())
        else:
            date_part = file[:-5]
            with open(filepath, 'r') as prfile:
                data = json.loads(prfile.read())
        if data:
            (self.fields_map, self.profiles) = data
            try:
                dtime = datetime.datetime.fromisoformat(date_part)
                self.datetime_str = dtime.strftime("%b %d %Y %I:%M%p")
            except ValueError:
                self.datetime_str = 'Unknown'


class EmptyProfile(ProfileData):
    """
    Used when there isn't actually any profile data.
    """

    def __init__(self):
        ProfileData.__init__(self)


class Results:
    def __init__(self,
                 reference_data: ProfileData,
                 current_data: ProfileData):
        """
        Save results of compare and method to save data after successful delivery
        :param reference_data: Previous data
        :param current_data: Current data
        """
        self.reference_data: ProfileData = reference_data
        self.current_data: ProfileData = current_data

    def get_diffs(self) -> List[Tuple[str, List[Tuple[str, List[str], List[str]]]]]:
        """
        Return result of compare
        :return: The list of profile differences
        """
        all_fields = dict(self.reference_data.get_fields_map())
        all_fields.update(self.current_data.get_fields_map())
        joined_values = join_dicts(self.reference_data.get_profile_fields(),
                                   self.current_data.get_profile_fields())

        diffs = profile_compare(joined_values, all_fields)
        # If there are multiple values in a bucket they're in random order
        # Sort them. Alpha order is nice for the report, but mostly
        # this makes repeatable unit test possible.
        for person, fields in diffs:
            for field in fields:
                if field[1]:
                    field[1].sort()
                if field[2]:
                    field[2].sort()
        return diffs

    def save_data(self) -> None:
        """
        Save recent data on successful send of email.
        :return: None
        """
        self.current_data.save_data()

    @property
    def header(self) -> str:
        """
        Return date range of differences as string.
        :return:
        """
        return (f'{self.reference_data.get_datetime()} thru '
                f'{self.current_data.get_datetime()}')


def _get_previous_files(args: argparse.Namespace) -> List[str]:
    if os.path.exists(args.data):
        return [f for f in os.listdir(args.data)
                if f.endswith(('.json', '.json.gz'))]
    else:
        return []


def _generate_diffs(args,
                    breeze_api: breeze.BreezeApi = None) -> Results:
    if args.reference_data:
        # Special case. Run current data against an existing reference
        # and don't save the results
        reference_data = ProfileFromFile(args.reference_data)
        current_data = ProfileFromBreeze(args, breeze_api, do_save=False)
    else:
        prev_saved = _get_previous_files(args)
        prev_saved.sort(reverse=True)
        if args.replay:
            # Generate report from the two previous runs
            if len(prev_saved) < 2:
                # Need two files for replay
                sys.exit(f'Need at two previous runs in {args.data} for replay')
            reference_data = ProfileFromFile(os.path.join(args.data,
                                                          prev_saved[1]))
            current_data = ProfileFromFile(os.path.join(args.data,
                                                        prev_saved[0]))
        else:
            if not prev_saved:
                logging.warning('No previous data found. This will be big!')
                reference_data = EmptyProfile()
            else:
                reference_data = ProfileFromFile(os.path.join(args.data,
                                                              prev_saved[0]))
            current_data = ProfileFromBreeze(args, breeze_api)

    return Results(reference_data, current_data)


def _verify_directory(args: argparse.Namespace, create=False):
    if os.path.isdir(args.data):
        if os.access(args.data, os.W_OK | os.R_OK):
            return
        msg = f'{args.data} exists but isn\'t writable'
    elif create:
        Path(args.data).mkdir(exist_ok=True, parents=True)
        logging.info(f'Creating directory {args.data}')
        return
    else:
        msg = f'Directory {args.data} doesn\'t exist'

    logging.error(msg)
    sys.exit(msg)


def _list_domains(args):
    domains = known_domains(overrides=args.email_servers)
    print("Known email domains:")
    for domain, text in domains.items():
        print(f'   {domain}: {text}')


def main(breeze_api: breeze.BreezeApi = None,
         email_sender: MailSender = None):
    """
    Email a report of Breeze profile changes
    :param breeze_api: (For testing) pre-made Breeze API
    :param email_sender: (For testing) email sender
    :return: None
    """

    # First, figure out what we'll be doing
    parser = argparse.ArgumentParser('Generate report of recent Breeze changes')
    parser.add_argument('--from',
                        '-f',
                        dest='sender',
                        metavar='sender',
                        help='sending email address (required in most cases)')
    parser.add_argument('--to',
                        '-t',
                        metavar='recipient(s)',
                        help='public recipients, comma separated')
    parser.add_argument('--bcc',
                        '-b',
                        metavar='<blind recipient(s)>',
                        help='hidden recipients, comma separated')
    parser.add_argument('--data',
                        '-d',
                        metavar='<data directory>',
                        help=f'directory with history, default: {DEFAULT_DATA_DIR}',
                        default=DEFAULT_DATA_DIR)
    parser.add_argument('--reference_data',
                        metavar='<reference data file>',
                        help='explicit previous data file to compare against')
    parser.add_argument('--format',
                        metavar='<report format>',
                        help='form of report (html, text, or csv)',
                        choices=['html', 'text', 'csv'],
                        default='html')
    parser.add_argument('--columns',
                        metavar='<column widths>',
                        help='column widths for format=text, default: ' +
                             DEFAULT_COLUMN_WIDTHS,
                        default=DEFAULT_COLUMN_WIDTHS)
    parser.add_argument('--breeze_creds',
                        metavar='<Breeze credential file>',
                        help='file with Breeze credentials if not standard',
                        default=None)
    parser.add_argument('--email_creds',
                        metavar='<email credential file>',
                        help='file with email passwords if not standard')
    parser.add_argument('--email_servers',
                        metavar='<email domain specification file>',
                        help='file with email domain configuration if not standard')
    parser.add_argument('--replay',
                        help='send differences between last two snapshots',
                        default=False,
                        action='store_true')
    parser.add_argument('--list_directories',
                        default=False,
                        action='store_true',
                        help='list data and configuration directories and exit')
    parser.add_argument('--initialize',
                        default=False,
                        action='store_true',
                        help="initialize from Breeze before first use "
                             "without sending report")
    parser.add_argument('--list_domains',
                        default=False,
                        action='store_true',
                        help='Print known email domains and exit')
    parser.add_argument('--retain_days',
                        type=int,
                        metavar='<days to retain>',
                        default=0,
                        help='<Number of days to retain old data')
    parser.add_argument('--logfile',
                        default=DEFAULT_LOG_FILE,
                        metavar='Log file',
                        help='File for application logs')
    parser.add_argument('--log_level',
                        default='info',
                        metavar='Log level',
                        help='Logging level, default=info')

    args = parser.parse_args()

    if args.logfile == 'stdout':
        out = {'stream': sys.stdout}
    elif args.logfile == 'stderr':
        out = {'stream': sys.stderr}
    else:
        out = {'filename': args.logfile}
        directory, file = os.path.split(args.logfile)
        Path(directory).mkdir(exist_ok=True, parents=True)

    logging.basicConfig(
        format="%(asctime)s [%(levelname)s] %(message)s",
        level=getattr(logging, args.log_level.upper()),
        **out
    )

    logging.info(f'Running {" ".join(sys.argv)}')

    if args.list_domains:
        _list_domains(args)
        sys.exit(0)

    if args.list_directories:
        # Making it easy to note where users should look for files on this platform.
        mailsender_configs = configured_mail_sender.config_file_list()
        print('configured_mail_sender configuration files:')
        for f in mailsender_configs:
            print(f'\t{f}')
        breeze_config = breeze.config_file_list(overrides=args.breeze_creds)
        print('breeze_chms_api configuration files:')
        for f in breeze_config:
            print(f'\t{f}')
        print(f'{APPLICATION_NAME} data directory:\n\t{args.data}')
        print(f'log output\n\t{args.logfile}')
        sys.exit(0)

    if args.initialize:
        _verify_directory(args, create=True)
        if _get_previous_files(args):
            # Shouldn't initialize if there's already data
            sys.exit(f'Remove files from {args.data} before initializing')
        # Create first-time reference data to prevent monster first report
        current_data = ProfileFromBreeze(args, breeze_api)
        current_data.save_data()
        msg = f'Initial data saved to {current_data.save_file_name}'
        logging.info(msg)
        print(msg)
        sys.exit(0)

    if not args.sender:
        sys.exit('--from=sender is required')

    if not args.to and not args.bcc:
        sys.exit('Either -t or -b is required')

    _verify_directory(args)

    results = _generate_diffs(args, breeze_api)

    widths = [int(w) for w in args.columns.split(',')]
    names = ['Field', 'Old', 'New']
    column_specs = [ColSpec(names[i], width=widths[i]) for i in range(3)]
    msg = MIMEMultipart()
    msg['Subject'] = 'Breeze profile change report'
    if args.to:
        msg['To'] = args.to
    if args.bcc:
        msg['Bcc'] = args.bcc
    diffs = results.get_diffs()
    if diffs:
        msg.attach(MIMEText(f'Breeze changes for {results.header}'))
        if args.format == 'html':
            formatter = _HTMLFormatter(column_specs)
        elif args.format == 'csv':
            formatter = _CSVFormatter(column_specs)
        else:
            args.format = 'plain'
            formatter = _TextFormatter(column_specs)
        output = StringIO()
        formatter.format_table(diffs, output)
        output.seek(0)
        txt = output.read()
        if args.format == 'csv':
            part = MIMEText(txt, 'csv', 'utf-8')
            # part.set_payload(txt)
            part.add_header('Content-Disposition',
                            "attachment; filename=profile_changes.csv")
            msg.attach(part)
        else:
            msg.attach(MIMEText(txt, args.format))
    else:
        msg.attach(MIMEText(f'No changes found between {results.header}'))

    sender = email_sender if email_sender else \
        create_sender(args.sender,
                      creds_file=args.email_creds,
                      overrides=args.email_servers)
    sender.send_message(msg)

    results.save_data()


if __name__ == "__main__":
    main()
